<?php

namespace Oreolek\Downloader;

use GuzzleHttp\Client as GuzzleClient;
use GuzzleHttp\HandlerStack;
use GuzzleHttp\Handler\CurlHandler;
use Illuminate\Support\Facades\Cache;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Storage;
use Symfony\Component\DomCrawler\Crawler;
use \Campo\UserAgent;
use Goutte\Client;
use Nesk\Puphpeteer\Puppeteer;
use Nesk\Rialto\Data\JsFunction;

class Downloader {
    protected $guzzle;
    protected $goutte;
    protected $proxy;
    public const PROXY_ELITE = 1;
    public const PROXY_GOOD = 2;
    public const PROXY_ANYTHING = 3;
    public $proxyLevel = self::PROXY_ANYTHING;
    protected $browser;
    protected $browserPage;

    /**
     * Useragent string.
     *
     * @var string
     */
    public $useragent;

    public function __construct()
    {
        $this->changeUserAgent();
        $this->recreateClients();
    }

    public function getGoutte() {
        return $this->goutte;
    }
    public function getGuzzle()
    {
        return $this->guzzle;
    }

    public function setGuzzle($guzzle)
    {
        $this->guzzle = $guzzle;
    }

    /**
     * Default Guzzle config.
     */
    protected $defaultConfig = [
        'debug' => false,
        'timeout' => 10,
        'cookies' => true, // use a shared client cookie jar
        'verify' => false, // verify https
        'connect_timeout' => 10,
        'headers' => [
            'Accept-Encoding' => 'gzip, deflate',
            'Accept-Language' => 'en-US,en;q=0.5',
            'Connection' => 'keep-alive',
            'Cache-Control' => 'no-cache',
            'Content-Encoding' => 'gzip, deflate',
        ],
        'decode_content' => true,
    ];

    public function setProxy($proxy)
    {
        $this->defaultConfig['proxy'] = $proxy;
        $this->recreateClients();
        $this->proxy = $proxy;
    }

    /**
     * Unset and create guzzle client with new config.
     */
    public function recreateClients($config = []) {
        $handler = new CurlHandler();
        $stack = HandlerStack::create($handler);
        unset($this->guzzle);
        $guzzleconfig = array_merge($config, $this->defaultConfig);
        $this->guzzle = new GuzzleClient($guzzleconfig);
        $this->goutte = new Client();
        $this->goutte->setClient($this->guzzle);
    }

    /**
     * Clears the proxy.
     */
    public function clearProxy()
    {
        unset($this->defaultConfig['proxy']);
        $this->proxy = NULL;
        $this->recreateClients();
    }

    public function getProxy() {
        return $this->proxy;
    }

    public function request($method, $url, $options = []) {
        $response = NULL;
        for ($i = 0; $i < 3; $i++) {
            try {
                $response = $this->guzzle->request($method, $url, $options);
                break;
            } catch (\Throwable $e) {
                Log::error($e->getMessage());
                $this->changeProxy(true);
            }
        }
        if (!isset($response)) {
            Log::warning('No data.');
            return NULL;
        }
        return $response;
    }

    /**
     * Change current proxy.
     * Gets a random proxy address and tests it on a known URL.
     *
     * @param bool $force should we force a new proxy if CHANGE_PROXY = false
     */
    public function changeProxy($force = false) {
        unset($this->guzzle);
        $this->changeUserAgent();
        $this->proxy = NULL;
        if (env('USE_PROXY') !== false) {
            $this->guzzle = new GuzzleClient($this->defaultConfig);
            $proxies = $this->updateProxies();
            $testurl = env('PROXYTEST_URL');
            $proxy = Cache::get('proxy');
            // if we have a cached proxy
            // and we want to use the cached proxy
            // and we don't want to force refresh
            if (!empty($proxy) && env('CHANGE_PROXY', true) === false && $force === false) {
                $this->setProxy($proxy);
            } else {
                foreach ($proxies as $index => $proxy) {
                    $proxy = trim($proxy);
                    if ($proxy === '') {
                        continue;
                    }
                    if ($this->proxyLevel === self::PROXY_ANYTHING) {
                        Log::debug('Set '.$proxy.' as default.');
                        $this->setProxy($proxy);
                        Cache::put('proxy', $proxy, now()->addHours(6));
                        break;
                    }
                    Log::debug('Checking '.$proxy.'...');
                    try {
                        $response = $this->guzzle->request('GET', $testurl, [
                            'proxy' => $proxy,
                            'params' => [
                                'ip' => $proxy
                            ]
                        ]);
                        $data = (string) $response->getBody();
                        $data = json_decode($data);
                        if ($data === NULL) {
                            Log::debug('Not a JSON response; skipping this proxy.');
                            // not a JSON response, skipping
                            unset($proxies[$index]);
                            continue;
                        }
                        if (
                            $response->getStatusCode() === 200 &&
                            (
                                $this->proxyLevel === self::PROXY_GOOD ||
                                (
                                    $this->proxyLevel === self::PROXY_ELITE &&
                                    isset($data->transparent) &&
                                    $data->transparent === false
                                )
                            )
                        ) {
                            Log::debug('Set '.$proxy.' as default.');
                            $this->setProxy($proxy);
                            Cache::put('proxy', $proxy, now()->addHours(6));
                            break;
                        } else {
                            Log::debug('Proxy is not transparent. Skipping.');
                        }
                    } catch (\Throwable $e) {
                        Log::error($e->getMessage());
                        unset($proxies[$index]);
                    }
                }
                if ($this->proxy === NULL) {
                    // we still did not set a proxy
                    Log::error('No proxies selected. Cannot continue. Abort.');
                    Cache::forget('proxies');
                    Cache::forget('proxy');
                    throw new \Exception('No proxies');
                }
                // we marked some proxies as faulty, cache this
                Cache::put('proxies', $proxies, now()->addDays(1));
            }
        }
    }

    public function changeUserAgent() {
        $useragent = UserAgent::random(['device_type' => 'Desktop']);
        $this->defaultConfig['headers']['User-Agent'] = $useragent;
        $this->useragent = $useragent;
        $this->recreateClients();
    }

    /**
     * Update the list of proxies if it's outdated and return 100 random URLs.
     * Cache the list for 6 hours.
     *
     * @return array
     */
    public function updateProxies()
    {
        if (env('USE_PROXY') === false) {
            return [];
        }
        if (!empty(env('MANUAL_PROXY'))) {
            $proxies = env('MANUAL_PROXY');
            if (!is_array($proxies)) {
                $proxies = [$proxies];
            }
        } else {
            $proxies = Cache::get('proxies', []);
            if (empty($proxies)) {
                $this->clearProxy();
                try {
                    $response = $this->guzzle->request('GET', 'http://api.foxtools.ru/v2/Proxy.txt?uptime=1');
                } catch (\Throwable $e) {
                }
                if (
                    isset($response) &&
                    (
                        $response->getStatusCode() === 200 ||
                        $response->getStatusCode() === 302
                    )
                ) {
                    $response = (string) $response->getBody();
                    $line = strtok($response, PHP_EOL);
                    $proxies[] = trim($line);
                    while ($line !== FALSE) {
                        $line = strtok(PHP_EOL);
                        $proxies[] = trim($line);
                    }
                    strtok('', '');
                    unset($response);
                    unset($line);
                }

                for ($i = 1; $i <10; $i++) {
                    try {
                        $crawler = $this->getHTML('https://proxy-list.org/english/index.php?p='.$i);
                        $newProxies = $crawler
                            ->filter('li.proxy script')
                            ->each(function(Crawler $node, $i) {
                                $html = $node->text();
                                $info = explode("'", $html);
                                $info = trim(base64_decode($info[1]));
                                return $info;
                            });
                        $proxies = array_merge($proxies, $newProxies);
                    } catch (\Throwable $e) {
                    }
                }

                try {
                    $urls = [
                        'http://www.gatherproxy.com',
                        'http://www.gatherproxy.com/sockslist'
                    ];
                    foreach ($urls as $url) {
                        $crawler = $this->getHTML($url);
                        $newProxies = $crawler
                            ->filter('script')
                            ->each(function(Crawler $node, $i) {
                                $text = trim($node->text());
                                if (strpos($text, 'insertPrx') === FALSE) {
                                    return false;
                                }
                                $text = str_replace('gp.insertPrx(', '', $text);
                                $text = str_replace(');', '', $text);
                                $text = json_decode($text);
                                $ip = $text->PROXY_IP;
                                $port = hexdec($text->PROXY_PORT);
                                return $ip.':'.$port;
                            });
                        if (!empty($newProxies)) {
                            $newProxies = array_unique($newProxies);
                            if (($key = array_search(FALSE, $newProxies)) !== false) {
                                unset($newProxies[$key]);
                            }
                            $proxies = array_merge($proxies, $newProxies);
                        }
                    }
                } catch (\Throwable $e) {
                }

                try {
                    $crawler = $this->getHTML('https://www.sslproxies.org');
                    $newProxies = $crawler
                        ->filter('#proxylisttable td:first-child')
                        ->each(function(Crawler $node, $i) {
                            return $node->text();
                        });
                    $proxies = array_merge($proxies, $newProxies);
                } catch (\Throwable $e) {
                }

                // 20 premium proxies
                try {
                    $response = $this->guzzle->request('GET', 'http://pubproxy.com/api/proxy?limit=20&post=true&https=true&format=json&country=US,CA&api='.env('PUBPROXY_API'));
                } catch (\Throwable $e) {
                    Log::error($e->getMessage());
                }
                if (
                    isset($response) &&
                    (
                        $response->getStatusCode() === 200 ||
                        $response->getStatusCode() === 302
                    )
                ) {
                    $response = json_decode((string) $response->getBody());
                    if (is_object($response)) {
                        $response = $response->data;
                        foreach ($response as $proxy) {
                            $proxies[] = $proxy->ipPort;
                        }
                    }
                }
                $proxies = array_unique($proxies);

                Log::debug('Loaded a list of '.count($proxies).' proxies.');
                Cache::put('proxies', $proxies, now()->addDays(1));
            }
        }
        shuffle($proxies);
        return ($proxies);
    }

    /**
     * Downloads and decodes JSON data.
     *
     * @param string $url request URL
     * @param string $method request method
     * @param array $options request options
     * @return object|NULL decoded data or NULL if request wasn't successful
     */
    public function getJSON($url, $method = 'GET', $options = []) {
        $url = trim($url);
        if (preg_match('/^file:\/\//', $url)) {
            $url = str_replace('file://', '', $url);
            $response = file_get_contents($url);
            if ($response) {
                return json_decode($response);
            } else {
                return NULL;
            }
        }
        $errors = 0;
        if (!isset($options['headers'])) {
            $options['headers'] = [];
        }
        if (!isset($options['headers']['Accept'])) {
            $options['headers']['Accept'] = 'application/json';
        }
        while ($errors < 3) {
            try {
                $response = $this->guzzle->request($method, $url, $options);
                break;
            } catch (\GuzzleHttp\Exception\RequestException $e) {
                Log::error($e->getMessage());
                Log::debug($e->getTraceAsString());
                $errors++;
                $this->changeProxy(true);
            } catch (\Exception $e) {
                $errors++;
                Log::error($e->getMessage());
                Log::debug($e->getTraceAsString());
            }
        }
        if (!isset($response)) {
            Log::warning('No data.');
            return NULL;
        }
        $response = (string) $response->getBody();
        $response = json_decode($response);
        return $response;
    }

    public function getHTML($url, $method = 'GET', $options = []) {
        $url = trim($url);
        if (preg_match('/^file:\/\//', $url)) {
            $url = str_replace('file://', '', $url);
            $response = file_get_contents($url);
            if ($response) {
                return new Crawler($response);
            } else {
                return NULL;
            }
        }
        if (env('CACHE_REQUESTS', false) === true && Cache::has('request_'.$url)) {
            return new Crawler(Cache::get('request_'.$url));
        }
        $errors = 0;
        if (!isset($options['headers'])) {
            $options['headers'] = [];
        }
        if (!isset($options['headers']['Accept'])) {
            $options['headers']['Accept'] = 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8';
        }
        while ($errors < 3) {
            try {
                $response = $this->guzzle->request($method, $url, $options);
                break;
            } catch (\GuzzleHttp\Exception\ConnectException $e) {
                // Connection refused
                Log::error($e->getMessage());
                $errors++;
                $this->changeProxy(true);
            } catch (\GuzzleHttp\Exception\RequestException $e) {
                // 301 from proxy
                Log::error($e->getMessage());
                $errors++;
                $this->changeProxy(true);
            } catch (\Exception $e) {
                $errors++;
                Log::error($e->getMessage());
            }
        }
        if (isset($response)) {
            $response = (string) $response->getBody();
            if (env('CACHE_REQUESTS', false) === true && !empty($response)) {
                Cache::put('request_'.$url, $response, now()->addDays(7));
            }
            return new Crawler($response);
        }
        return NULL;
    }

    /**
     * Downloads and returns image data.
     *
     * @param string $url request URL
     * @param string $method request method
     * @param array $options request options
     * @return object|NULL decoded data or NULL if request wasn't successful
     */
    public function getImage($url, $method = 'GET', $options = []) {
        return $this->getFile($url, $method, $options);
    }

    public function getFile($url, $method = 'GET', $options = []) {
        try {
            $filename = $this->getFileResource($url, $method, $options);
        } catch (\Exception $e) {
            return false;
        }
        if (empty($filename)) {
            return FALSE;
        }
        $resource = fopen($filename, 'r');
        $contents = fread($resource, filesize($filename));
        fclose($resource);
        unlink($filename);
        return $contents;
    }

    /**
     * Less memory-intensive version.
     * But you have to manually read, close and delete the file.
     */
    public function getFileResource($url, $method = 'GET', $options = []) {
        $filename = sys_get_temp_dir().'/download_'.uniqid();
        Log::debug('Downloading '.$url.' into '.$filename);
        $cached = false;
        $local_url = pathinfo($url);
        $local_url = $local_url['basename'];
        $tempfile = fopen($filename, 'w');
        if ($tempfile === FALSE) {
            Log::error('Could not create temp file. Abort.'.PHP_EOL);
            return FALSE;
        }
        if (env('APP_DEBUG', false) && Storage::disk('local')->has($local_url)) {
            $contents = Storage::disk('local')->get($local_url);
            if (!empty($contents)) {
                fwrite($tempfile, $contents);
                fclose($tempfile);
                $cached = true;
            }
        }
        if ($cached === false) {
            $stream = \GuzzleHttp\Psr7\stream_for($tempfile);
            $errors = 0;
            $options['sink'] = $stream;
            while ($errors < 5) {
                try {
                    $this->request($method, $url, $options);
                    break;
                } catch ( \GuzzleHttp\Exception\RequestException $e) {
                    Log::error($e->getMessage());
                    // bad request
                    $this->changeProxy(true);
                    $errors++;
                    continue; // try again
                } catch( \GuzzleHttp\Exception\ServerException $e) {
                    // 5xx
                    Log::error('Server error. Skipping.'.PHP_EOL);
                    break;
                } catch( \GuzzleHttp\Exception\ClientException $e) {
                    Log::error($e->getMessage());
                    // 4xx
                    $this->changeProxy(true);
                    $errors++;
                    continue;
                }
            }
            if ($stream->getSize() === 0) {
                Log::error('Got an empty file. Abort download.'.PHP_EOL);
                // empty stream; skip this
                return false;
            }

            if (env('APP_DEBUG', false)) {
                Storage::disk('local')->put($local_url, $stream);
            }
            fclose($tempfile);
            $stream->close();
        }
        Log::debug('Downloaded '.$url);
        return $filename;
    }

    /**
     * Get startup options for Puppeteer browser.
     */
    protected function getBrowserOptions() {
        return [
            'args' => [
                '--no-sandbox',
                '--disable-setuid-sandbox',
                '--proxy-server='.$this->proxy,
            ],
            'log_node_console' => true,
            'log_browser_console' => true,
        ];
    }

    public function startBrowser() {
        if (is_null($this->browser)) {
            $puppeteer = new Puppeteer([
                'read_timeout' => 90,
                'idle_timeout' => 300, // the browser can stay inactive for 300 seconds
            ]);
            $this->browser = $puppeteer->launch($this->getBrowserOptions());
        }
        if (is_null($this->browserPage)) {
            $this->browserPage = $this->browser->newPage();
        }
    }

    public function newPage() {
        $this->browserPage = $this->browser->newPage();
    }

    public function browserGoto($url) {
        return $this->browserPage->tryCatch->goto($url);
    }

    public function browserHtml() {
        return $this->browserPage->content();
    }

    public function stopBrowser() {
        try {
            $this->browser->close();
        } catch(\Exception $e) {
        }
        $this->browser = NULL;
        $this->browserPage = NULL;
    }

    public function browser() {
        return $this->browser;
    }

    public function browserPage() {
        return $this->browserPage;
    }
}
